"""Customized exceptions."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator, Sequence


class DownloadError(Exception):
    """Exception raised when the requested data is not available on the server.

    Parameters
    ----------
    err : str
        Service error message.
    """

    def __init__(self, err: str, url: str | None = None) -> None:
        self.message = "Service returned the following error message:\n"
        if url is None:
            self.message += err
        else:
            self.message += f"URL: {url}\nERROR: {err}\n"
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class ServiceError(Exception):
    """Exception raised when the requested data is not available on the server.

    Parameters
    ----------
    err : str
        Service error message.
    """

    def __init__(self, err: str, url: str | None = None) -> None:
        self.message = "Service returned the following error message:\n"
        if url is None:
            self.message += err
        else:
            self.message += f"URL: {url}\nERROR: {err}\n"
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class InputValueError(Exception):
    """Exception raised for invalid input.

    Parameters
    ----------
    inp : str
        Name of the input parameter
    valid_inputs : tuple
        List of valid inputs
    given : str, optional
        The given input, defaults to None.
    """

    def __init__(
        self,
        inp: str,
        valid_inputs: Sequence[str | int] | Generator[str | int, None, None],
        given: str | int | None = None,
    ) -> None:
        if given is None:
            self.message = f"Given {inp} is invalid. Valid options are:\n"
        else:
            self.message = f"Given {inp} ({given}) is invalid. Valid options are:\n"
        self.message += "\n".join(str(i) for i in valid_inputs)
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class InputTypeError(Exception):
    """Exception raised when a function argument type is invalid.

    Parameters
    ----------
    arg : str
        Name of the function argument
    valid_type : str
        The valid type of the argument
    example : str, optional
        An example of a valid form of the argument, defaults to None.
    """

    def __init__(
        self,
        arg: str,
        valid_type: str,
        example: str | None = None,
    ) -> None:
        self.message = f"The {arg} argument should be of type {valid_type}"
        if example is not None:
            self.message += f":\n{example}"
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message



"""Tests for exceptions and requests."""

from __future__ import annotations

from urllib.parse import urlencode

import pytest
from aiohttp import InvalidURL

import async_retriever as ar
from async_retriever.exceptions import InputTypeError, InputValueError, ServiceError
from async_retriever.streaming import DownloadError


@pytest.fixture
def url_kwds():
    stations = ["01646500", "08072300", "11073495"]
    url = "https://waterservices.usgs.gov/nwis/site"
    return zip(
        *((url, {"params": {"format": "rdb", "sites": s, "siteStatus": "all"}}) for s in stations),
    )


def test_invalid_method(url_kwds):
    urls, kwds = url_kwds
    with pytest.raises(InputValueError) as ex:
        _ = ar.retrieve(urls, "text", request_kwds=kwds, request_method="getter")
    assert "get" in str(ex.value)


def test_delete_invalid_method(url_kwds):
    urls, _ = url_kwds
    with pytest.raises(InputValueError) as ex:
        ar.delete_url_cache(urls[0], request_method="getter")
    assert "get" in str(ex.value)


def test_invalid_read(url_kwds):
    urls, kwds = url_kwds
    with pytest.raises(InputValueError) as ex:
        _ = ar.retrieve(urls, "texts", request_kwds=kwds)
    assert "read" in str(ex.value)


def test_invalid_url(url_kwds):
    urls, kwds = url_kwds
    with pytest.raises(InputTypeError) as ex:
        _ = ar.retrieve(urls[0], "text", request_kwds=kwds)
    assert "list of str" in str(ex.value)


def test_invalid_link():
    urls = ["dead.link.com"]
    with pytest.raises(InvalidURL) as ex:
        _ = ar.retrieve(urls, "text")
    assert "dead.link.com" in str(ex.value)


def test_invalid_length(url_kwds):
    urls, kwds = url_kwds
    with pytest.raises(InputTypeError) as ex:
        _ = ar.retrieve(urls * 2, "text", request_kwds=kwds)
    assert "the same size" in str(ex.value)


def test_invalid_kwds(url_kwds):
    urls, kwds = url_kwds
    kwds = [{"paramss": v} for kw in kwds for _, v in kw.items()]
    with pytest.raises(InputValueError) as ex:
        _ = ar.retrieve(urls, "text", request_kwds=kwds)
    assert "paramss" in str(ex.value)


def test_service_error():
    base_url = "https://api.water.usgs.gov/geoserver/wmadata/ows"
    params = {
        "bbox": "-96.1,28.7,-95.9,28.5,epsg:4326",
        "outputFormat": "application/json",
        "request": "GetFeature",
        "service": "wfs",
        "srsName": "epsg:4269",
        "typeName": "wmadata:nhdflowline_network",
        "version": "2.0.0",
    }
    url = f"{base_url}?{urlencode(params)}"
    with pytest.raises(ServiceError) as ex:
        _ = ar.retrieve([url], "json")
    assert "illegal bbox" in str(ex.value)

    with pytest.raises(DownloadError) as ex:
        _ = ar.stream_write([url], ["temp"])
    assert "illegal bbox" in str(ex.value)


def test_wrong_path_type():
    urls = ["https://freetestdata.com/wp-content/uploads/2021/09/Free_Test_Data_500KB_CSV-1.csv"]
    with pytest.raises(TypeError) as ex:
        _ = ar.stream_write(urls, "temp")
    assert "sequences" in str(ex.value)


def test_wrong_path_number():
    urls = ["https://freetestdata.com/wp-content/uploads/2021/09/Free_Test_Data_500KB_CSV-1.csv"]
    file_paths = ["temp"] * 2
    with pytest.raises(TypeError) as ex:
        _ = ar.stream_write(urls, file_paths)
    assert "same length" in str(ex.value)



"""Core async functions."""

from __future__ import annotations

import asyncio
import os
from typing import TYPE_CHECKING, Any, Literal, Union

import ujson as json
from aiohttp import ClientSession, TCPConnector
from aiohttp_client_cache import SQLiteBackend
from aiohttp_client_cache.session import CachedSession

from async_retriever import _utils as utils
from async_retriever._utils import BaseRetriever
from async_retriever.exceptions import InputValueError

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path
    from ssl import SSLContext

    from aiohttp.typedefs import StrOrURL

    Response = Union[
        "list[str]",
        "list[bytes]",
        "list[dict[str, Any]]",
        "list[list[dict[str, Any]]]",
    ]
EXPIRE_AFTER = 60 * 60 * 24 * 7  # 1 week
__all__ = [
    "delete_url_cache",
    "retrieve",
    "retrieve_binary",
    "retrieve_json",
    "retrieve_text",
]


def delete_url_cache(
    url: StrOrURL,
    request_method: Literal["get", "post"] = "get",
    cache_name: Path | str | None = None,
    **kwargs: dict[str, Any],
) -> None:
    """Delete cached response associated with ``url``, along with its history (if applicable).

    Parameters
    ----------
    url : str
        URL to be deleted from the cache
    request_method : str, optional
        HTTP request method to be deleted from the cache, defaults to ``GET``.
    cache_name : str, optional
        Path to a file for caching the session, defaults to
        ``./cache/aiohttp_cache.sqlite``.
    kwargs : dict, optional
        Keywords to pass to the ``cache.delete_url()``.
    """
    loop, new_loop = utils.get_event_loop()

    valid_methods = ["get", "post"]
    if request_method.lower() not in valid_methods:
        raise InputValueError("method", valid_methods)

    loop.run_until_complete(
        utils.delete_url(url, request_method, utils.create_cachefile(cache_name), **kwargs)
    )
    if new_loop:
        loop.close()


async def _session_with_cache(
    url_kwds: zip[tuple[int, StrOrURL, dict[str, Any]]],
    read: Literal["read", "text", "json"],
    r_kwds: dict[str, Any],
    request_method: Literal["get", "post"],
    cache_name: Path,
    timeout: int = 5,
    expire_after: int = EXPIRE_AFTER,
    ssl: SSLContext | bool = True,
    raise_status: bool = True,
    limit_per_host: int = 5,
) -> (
    list[tuple[int, str | None]]
    | list[tuple[int, bytes | None]]
    | list[tuple[int, dict[str, Any] | None]]
    | list[tuple[int, list[dict[str, Any]] | None]]
):
    """Create an async session for sending requests.

    Parameters
    ----------
    url_kwds : list of tuples of urls and payloads
        A list of URLs or URLs with their payloads to be retrieved.
    read : str
        The method for returning the request; ``read`` (bytes), ``json``, and ``text``.
    r_kwds : dict
        Keywords to pass to the response read function. ``{"content_type": None}`` if read
        is ``json`` else it's empty.
    request_method : str
        The request type; ``GET`` or ``POST``.
    cache_name : str
        Path to a file for caching the session, defaults to
        ``./cache/aiohttp_cache.sqlite``.
    timeout : int, optional
        Requests timeout in seconds, defaults to 5.
    expire_after : int, optional
        Expiration time for the cache in seconds, defaults to 2592000 (one week).
    ssl : bool or SSLContext, optional
        SSLContext to use for the connection, defaults to None. Set to ``False`` to disable
        SSL certification verification.
    raise_status : bool, optional
        Raise an exception if the response status is not 200. If
        ``False`` return ``None``. Defaults to ``True``.
    limit_per_host : int, optional
        Maximum number of simultaneous connections per host, defaults to 5.

    Returns
    -------
    asyncio.gather
        An async gather function
    """
    if expire_after == EXPIRE_AFTER:
        expire_after = int(os.getenv("HYRIVER_CACHE_EXPIRE", EXPIRE_AFTER))
    cache = SQLiteBackend(
        cache_name=str(cache_name),
        expire_after=expire_after,
        allowed_methods=("GET", "POST"),
        timeout=timeout,
        fast_save=True,
    )
    async with CachedSession(
        json_serialize=json.dumps,
        cache=cache,
        connector=TCPConnector(ssl=ssl, limit_per_host=limit_per_host),
        trust_env=True,
    ) as session:
        request_func = getattr(session, request_method.lower())
        tasks = [
            utils.retriever(uid, url, kwds, request_func, read, r_kwds, raise_status)
            for uid, url, kwds in url_kwds
        ]
        return await asyncio.gather(*tasks)  # pyright: ignore[reportReturnType]


async def _session_without_cache(
    url_kwds: zip[tuple[int, StrOrURL, dict[str, Any]]],
    read: Literal["read", "text", "json"],
    r_kwds: dict[str, Any],
    request_method: Literal["get", "post"],
    ssl: SSLContext | bool = True,
    raise_status: bool = True,
    limit_per_host: int = 5,
) -> (
    list[tuple[int, str | None]]
    | list[tuple[int, bytes | None]]
    | list[tuple[int, dict[str, Any] | None]]
    | list[tuple[int, list[dict[str, Any]] | None]]
):
    """Create an async session for sending requests.

    Parameters
    ----------
    url_kwds : list of tuples of urls and payloads
        A list of URLs or URLs with their payloads to be retrieved.
    read : str
        The method for returning the request; ``read`` (bytes), ``json``, and ``text``.
    r_kwds : dict
        Keywords to pass to the response read function. ``{"content_type": None}`` if read
        is ``json`` else it's empty.
    request_method : str
        The request type; ``GET`` or ``POST``.
    ssl : bool or SSLContext, optional
        SSLContext to use for the connection, defaults to None. Set to ``False`` to disable
        SSL certification verification.
    raise_status : bool, optional
        Raise an exception if the response status is not 200. If
        ``False`` return ``None``. Defaults to ``True``.
    limit_per_host : int, optional
        Maximum number of simultaneous connections per host, defaults to 5.

    Returns
    -------
    asyncio.gather
        An async gather function
    """
    async with ClientSession(
        json_serialize=json.dumps,
        trust_env=True,
        connector=TCPConnector(ssl=ssl, limit_per_host=limit_per_host),
    ) as session:
        request_func = getattr(session, request_method.lower())
        tasks = [
            utils.retriever(uid, url, kwds, request_func, read, r_kwds, raise_status)
            for uid, url, kwds in url_kwds
        ]
        return await asyncio.gather(*tasks)  # pyright: ignore[reportReturnType]


def retrieve(
    urls: Sequence[StrOrURL],
    read_method: Literal["text", "json", "binary"],
    request_kwds: Sequence[dict[str, Any]] | None = None,
    request_method: Literal["get", "post"] = "get",
    limit_per_host: int = 5,
    cache_name: Path | str | None = None,
    timeout: int = 5,
    expire_after: int = EXPIRE_AFTER,
    ssl: SSLContext | bool = True,
    disable: bool = False,
    raise_status: bool = True,
) -> list[str | bytes | dict[str, Any] | list[dict[str, Any]] | None]:
    r"""Send async requests.

    Parameters
    ----------
    urls : list of str
        List of URLs.
    read_method : str
        Method for returning the request; ``binary``, ``json``, and ``text``.
    request_kwds : list of dict, optional
        List of requests keywords corresponding to input URLs (1 on 1 mapping),
        defaults to ``None``. For example, ``[{"params": {...}, "headers": {...}}, ...]``.
    request_method : str, optional
        Request type; ``GET`` (``get``) or ``POST`` (``post``). Defaults to ``GET``.
    limit_per_host : int, optional
        Maximum number of simultaneous connections per host, defaults to 5.
    cache_name : str, optional
        Path to a file for caching the session, defaults to ``./cache/aiohttp_cache.sqlite``.
    timeout : int, optional
        Requests timeout in seconds, defaults to 5.
    expire_after : int, optional
        Expiration time for response caching in seconds, defaults to 2592000 (one week).
    ssl : bool or SSLContext, optional
        SSLContext to use for the connection, defaults to None. Set to False to disable
        SSL certification verification.
    disable : bool, optional
        If ``True`` temporarily disable caching requests and get new responses
        from the server, defaults to False.
    raise_status : bool, optional
        Raise an exception if the response status is not 200. If
        ``False`` return ``None``. Defaults to ``True``.

    Returns
    -------
    list
        List of responses in the order of input URLs.

    Examples
    --------
    >>> import async_retriever as ar
    >>> stations = ["01646500", "08072300", "11073495"]
    >>> url = "https://waterservices.usgs.gov/nwis/site"
    >>> urls, kwds = zip(
    ...     *[
    ...         (url, {"params": {"format": "rdb", "sites": s, "siteStatus": "all"}})
    ...         for s in stations
    ...     ]
    ... )
    >>> resp = ar.retrieve(urls, "text", request_kwds=kwds)
    >>> resp[0].split("\n")[-2].split("\t")[1]
    '01646500'
    """
    inp = BaseRetriever(
        urls,
        read_method=read_method,
        request_kwds=request_kwds,
        request_method=request_method,
        cache_name=cache_name,
        ssl=ssl,
    )

    if not disable:
        disable = os.getenv("HYRIVER_CACHE_DISABLE", "false").lower() == "true"

    loop, new_loop = utils.get_event_loop()
    if disable:
        results = loop.run_until_complete(
            _session_without_cache(
                inp.url_kwds,
                inp.read_method,
                inp.r_kwds,
                inp.request_method,
                ssl,
                raise_status,
                limit_per_host,
            )
        )
    else:
        results = loop.run_until_complete(
            _session_with_cache(
                inp.url_kwds,
                inp.read_method,
                inp.r_kwds,
                inp.request_method,
                inp.cache_name,
                timeout,
                expire_after,
                ssl,
                raise_status,
                limit_per_host,
            )
        )
    if new_loop:
        loop.close()
    return [r for _, r in sorted(results)]


def retrieve_text(
    urls: Sequence[StrOrURL],
    request_kwds: Sequence[dict[str, Any]] | None = None,
    request_method: Literal["get", "post"] = "get",
    limit_per_host: int = 5,
    cache_name: Path | str | None = None,
    timeout: int = 5,
    expire_after: int = EXPIRE_AFTER,
    ssl: SSLContext | bool = True,
    disable: bool = False,
    raise_status: bool = True,
) -> list[str | None]:
    r"""Send async requests and get the response as ``text``.

    Parameters
    ----------
    urls : list of str
        List of URLs.
    request_kwds : list of dict, optional
        List of requests keywords corresponding to input URLs (1 on 1 mapping),
        defaults to ``None``. For example, ``[{"params": {...}, "headers": {...}}, ...]``.
    request_method : str, optional
        Request type; ``GET`` (``get``) or ``POST`` (``post``). Defaults to ``GET``.
    limit_per_host : int, optional
        Maximum number of simultaneous connections per host, defaults to 5.
    cache_name : str, optional
        Path to a file for caching the session, defaults to ``./cache/aiohttp_cache.sqlite``.
    timeout : int, optional
        Requests timeout in seconds in seconds, defaults to 5.
    expire_after : int, optional
        Expiration time for response caching in seconds, defaults to 2592000 (one week).
    ssl : bool or SSLContext, optional
        SSLContext to use for the connection, defaults to None. Set to False to disable
        SSL certification verification.
    disable : bool, optional
        If ``True`` temporarily disable caching requests and get new responses
        from the server, defaults to False.
    raise_status : bool, optional
        Raise an exception if the response status is not 200. If
        ``False`` return ``None``. Defaults to ``True``.

    Returns
    -------
    list
        List of responses in the order of input URLs.

    Examples
    --------
    >>> import async_retriever as ar
    >>> stations = ["01646500", "08072300", "11073495"]
    >>> url = "https://waterservices.usgs.gov/nwis/site"
    >>> urls, kwds = zip(
    ...     *[
    ...         (url, {"params": {"format": "rdb", "sites": s, "siteStatus": "all"}})
    ...         for s in stations
    ...     ]
    ... )
    >>> resp = ar.retrieve_text(urls, kwds)
    >>> resp[0].split("\n")[-2].split("\t")[1]
    '01646500'
    """
    return retrieve(  # pyright: ignore[reportReturnType]
        urls,
        "text",
        request_kwds,
        request_method,
        limit_per_host,
        cache_name,
        timeout,
        expire_after,
        ssl,
        disable,
        raise_status,
    )


def retrieve_json(
    urls: Sequence[StrOrURL],
    request_kwds: Sequence[dict[str, Any]] | None = None,
    request_method: Literal["get", "post"] = "get",
    limit_per_host: int = 5,
    cache_name: Path | str | None = None,
    timeout: int = 5,
    expire_after: int = EXPIRE_AFTER,
    ssl: SSLContext | bool = True,
    disable: bool = False,
    raise_status: bool = True,
) -> list[dict[str, Any] | None] | list[list[dict[str, Any]] | None]:
    r"""Send async requests and get the response as ``json``.

    Parameters
    ----------
    urls : list of str
        List of URLs.
    request_kwds : list of dict, optional
        List of requests keywords corresponding to input URLs (1 on 1 mapping),
        defaults to ``None``. For example, ``[{"params": {...}, "headers": {...}}, ...]``.
    request_method : str, optional
        Request type; ``GET`` (``get``) or ``POST`` (``post``). Defaults to ``GET``.
    limit_per_host : int, optional
        Maximum number of simultaneous connections per host, defaults to 5.
    cache_name : str, optional
        Path to a file for caching the session, defaults to ``./cache/aiohttp_cache.sqlite``.
    timeout : int, optional
        Requests timeout in seconds, defaults to 5.
    expire_after : int, optional
        Expiration time for response caching in seconds, defaults to 2592000 (one week).
    ssl : bool or SSLContext, optional
        SSLContext to use for the connection, defaults to None. Set to False to disable
        SSL certification verification.
    disable : bool, optional
        If ``True`` temporarily disable caching requests and get new responses
        from the server, defaults to False.
    raise_status : bool, optional
        Raise an exception if the response status is not 200. If
        ``False`` return ``None``. Defaults to ``True``.

    Returns
    -------
    dict
        List of responses in the order of input URLs.

    Examples
    --------
    >>> import async_retriever as ar
    >>> urls = ["https://labs.waterdata.usgs.gov/api/nldi/linked-data/comid/position"]
    >>> kwds = [
    ...     {
    ...         "params": {
    ...             "f": "json",
    ...             "coords": "POINT(-68.325 45.0369)",
    ...         },
    ...     },
    ... ]
    >>> r = ar.retrieve_json(urls, kwds)
    >>> print(r[0]["features"][0]["properties"]["identifier"])
    2675320
    """
    return retrieve(  # pyright: ignore[reportReturnType]
        urls,
        "json",
        request_kwds,
        request_method,
        limit_per_host,
        cache_name,
        timeout,
        expire_after,
        ssl,
        disable,
        raise_status,
    )


def retrieve_binary(
    urls: Sequence[StrOrURL],
    request_kwds: Sequence[dict[str, Any]] | None = None,
    request_method: Literal["get", "post"] = "get",
    limit_per_host: int = 5,
    cache_name: Path | str | None = None,
    timeout: int = 5,
    expire_after: int = EXPIRE_AFTER,
    ssl: SSLContext | bool = True,
    disable: bool = False,
    raise_status: bool = True,
) -> list[bytes] | list[bytes | None]:
    r"""Send async requests and get the response as ``bytes``.

    Parameters
    ----------
    urls : list of str
        List of URLs.
    request_kwds : list of dict, optional
        List of requests keywords corresponding to input URLs (1 on 1 mapping),
        defaults to ``None``. For example, ``[{"params": {...}, "headers": {...}}, ...]``.
    request_method : str, optional
        Request type; ``GET`` (``get``) or ``POST`` (``post``). Defaults to ``GET``.
    limit_per_host : int, optional
        Maximum number of simultaneous connections per host, defaults to 5.
    cache_name : str, optional
        Path to a file for caching the session, defaults to ``./cache/aiohttp_cache.sqlite``.
    timeout : int, optional
        Requests timeout in seconds, defaults to 5.
    expire_after : int, optional
        Expiration time for response caching in seconds, defaults to 2592000 (one week).
    ssl : bool or SSLContext, optional
        SSLContext to use for the connection, defaults to None. Set to False to disable
        SSL certification verification.
    disable : bool, optional
        If ``True`` temporarily disable caching requests and get new responses
        from the server, defaults to False.
    raise_status : bool, optional
        Raise an exception if the response status is not 200. If
        ``False`` return ``None``. Defaults to ``True``.

    Returns
    -------
    bytes
        List of responses in the order of input URLs.
    """
    return retrieve(  # pyright: ignore[reportReturnType]
        urls,
        "binary",
        request_kwds,
        request_method,
        limit_per_host,
        cache_name,
        timeout,
        expire_after,
        ssl,
        disable,
        raise_status,
    )



"""Tests for the package."""

from __future__ import annotations

import asyncio
import io
import sys
import tempfile
from datetime import datetime
from pathlib import Path

from aiohttp_client_cache import SQLiteBackend

import async_retriever as ar

SMALL = 1e-3


async def check_url(url, method="GET", **kwargs):
    cache = SQLiteBackend(cache_name=Path("cache", "aiohttp_cache.sqlite"))
    return await cache.has_url(url, method, **kwargs)


def test_encoding():
    url = "https://api.epa.gov/StreamCat/streams/variable_info"
    resp = ar.retrieve_text([url])
    assert resp[0][:3] == "IND"


def test_disable_cache():
    url = "https://epqs.nationalmap.gov/v1/json"
    payload = {"params": {"x": -101, "y": 38, "units": "Meters"}}
    resp = ar.retrieve([url], "json", [payload], disable=True)
    elev = float(resp[0]["value"])
    assert abs(elev - 880.418) < SMALL
    assert not asyncio.run(check_url(url, params=payload["params"]))


def test_delete_url():
    url = "https://epqs.nationalmap.gov/v1/json"
    payload = {"params": {"x": -100, "y": 38, "units": "Meters"}}
    resp = ar.retrieve([url], "json", [payload])
    elev = float(resp[0]["value"])
    ar.delete_url_cache(url, params=payload["params"])
    assert abs(elev - 761.684) < SMALL
    assert not asyncio.run(check_url(url, params=payload["params"]))


def test_binary():
    west, south, east, north = (-69.77, 45.07, -69.31, 45.45)
    base_url = "https://thredds.daac.ornl.gov/thredds/ncss/ornldaac/1299"
    dates_itr = [(datetime(y, 1, 1), datetime(y, 1, 31)) for y in range(2000, 2005)]
    urls, kwds = zip(
        *(
            (
                f"{base_url}/MCD13.A{s.year}.unaccum.nc4",
                {
                    "params": {
                        "var": "NDVI",
                        "north": str(north),
                        "west": str(west),
                        "east": str(east),
                        "south": str(south),
                        "disableProjSubset": "on",
                        "horizStride": "1",
                        "time_start": s.strftime("%Y-%m-%dT%H:%M:%SZ"),
                        "time_end": e.strftime("%Y-%m-%dT%H:%M:%SZ"),
                        "timeStride": "1",
                        "addLatLon": "true",
                        "accept": "netcdf",
                    },
                },
            )
            for s, e in dates_itr
        ),
    )

    cache_name = "cache/tmp/aiohttp_cache.sqlite"
    r = ar.retrieve(urls, "binary", request_kwds=kwds, cache_name=cache_name, ssl=False)
    r_b = ar.retrieve_binary(urls, request_kwds=kwds, cache_name=cache_name, ssl=False)
    assert sys.getsizeof(r[0]) == sys.getsizeof(r_b[0]) == 986161


def test_json():
    urls = ["https://api.water.usgs.gov/nldi/linked-data/comid/position"]
    kwds = [
        {
            "params": {
                "f": "json",
                "coords": "POINT(-68.325 45.0369)",
            },
        },
    ]
    r = ar.retrieve(urls, "json", kwds)
    r_j = ar.retrieve_json(urls, kwds)
    r_id = r[0]["features"][0]["properties"]["identifier"]
    rj_id = r_j[0]["features"][0]["properties"]["identifier"]
    assert r_id == rj_id == "2675320"


def test_text_post():
    base = "https://waterservices.usgs.gov/nwis/site/?"
    station_id = "01646500"
    urls = [
        "&".join([base, "format=rdb", f"sites={','.join([station_id] * 20)}", "siteStatus=all"])
    ]

    r = ar.retrieve(urls, "text", request_method="POST")
    r_id = r[0].split("\n")[-2].split("\t")[1]

    r_t = ar.retrieve_text(urls, request_method="POST")
    rt_id = r_t[0].split("\n")[-2].split("\t")[1]

    assert r_id == rt_id == station_id


def test_stream():
    url = "https://freetestdata.com/wp-content/uploads/2021/09/Free_Test_Data_500KB_CSV-1.csv"
    with tempfile.NamedTemporaryFile(dir=".") as temp:
        ar.stream_write([url], [temp.name])
        assert Path(temp.name).stat().st_size == 512789


def test_stream_chunked():
    url = "https://freetestdata.com/wp-content/uploads/2021/09/Free_Test_Data_500KB_CSV-1.csv"
    with tempfile.NamedTemporaryFile(dir=".") as temp:
        ar.stream_write([url], [temp.name], chunk_size=5000)
        assert Path(temp.name).stat().st_size == 512789


def test_ordered_return():
    stations = ["11073495", "08072300", "01646500"]
    url = "https://waterservices.usgs.gov/nwis/site"
    urls, kwds = zip(
        *((url, {"params": {"format": "rdb", "sites": s, "siteStatus": "all"}}) for s in stations),
    )
    resp = ar.retrieve(urls, "text", request_kwds=kwds)
    assert [r.split("\n")[-2].split("\t")[1] for r in resp] == stations


def test_show_versions():
    f = io.StringIO()
    ar.show_versions(file=f)
    assert "SYS INFO" in f.getvalue()



"""Top-level package."""

from __future__ import annotations

import asyncio
import sys
from importlib.metadata import PackageNotFoundError, version

from async_retriever import exceptions
from async_retriever.async_retriever import (
    delete_url_cache,
    retrieve,
    retrieve_binary,
    retrieve_json,
    retrieve_text,
)
from async_retriever.print_versions import show_versions
from async_retriever.streaming import stream_write

try:
    __version__ = version("async_retriever")
except PackageNotFoundError:
    __version__ = "999"

if sys.platform == "win32":  # pragma: no cover
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

__all__ = [
    "__version__",
    "delete_url_cache",
    "exceptions",
    "retrieve",
    "retrieve_binary",
    "retrieve_json",
    "retrieve_text",
    "show_versions",
    "stream_write",
]



"""Core async functions."""

from __future__ import annotations

import asyncio
import os
from datetime import datetime
from inspect import signature
from itertools import repeat
from pathlib import Path
from ssl import PROTOCOL_TLS_CLIENT, SSLContext
from typing import TYPE_CHECKING, Any, Callable, Literal

import ujson as json
from aiohttp import ClientResponseError, ClientSession
from aiohttp_client_cache import SQLiteBackend

from async_retriever.exceptions import (
    InputTypeError,
    InputValueError,
    ServiceError,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from aiohttp.client import _RequestContextManager  # pyright: ignore[reportPrivateUsage]
    from aiohttp.typedefs import StrOrURL


def create_cachefile(db_name: str | Path | None = None) -> Path:
    """Create a cache folder in the current working directory."""
    if db_name is not None:
        fname = Path(db_name)
    else:
        fname = Path(os.getenv("HYRIVER_CACHE_NAME", Path("cache", "aiohttp_cache.sqlite")))
    # Delete cache files created before v0.4 (2023-03-01) since from then on
    # the expiration date of caches are set to 1 week.
    if fname.exists() and datetime.fromtimestamp(fname.stat().st_ctime) < datetime(2023, 3, 1):
        fname.unlink()
    fname.parent.mkdir(parents=True, exist_ok=True)
    return fname


async def retriever(
    uid: int,
    url: StrOrURL,
    s_kwds: dict[str, dict[str, Any]],
    session: Callable[[StrOrURL], _RequestContextManager],
    read_type: Literal["text", "read", "json"],
    r_kwds: dict[str, None],
    raise_status: bool,
) -> (
    tuple[int, str]
    | tuple[int, bytes]
    | tuple[int, dict[str, Any]]
    | tuple[int, list[dict[str, Any]]]
    | tuple[int, None]
):
    """Create an async request and return the response as binary.

    Parameters
    ----------
    uid : int
        ID of the URL for sorting after returning the results
    url : str
        URL to be retrieved
    s_kwds : dict
        Arguments to be passed to requests
    session : ClientSession
        A ClientSession for sending the request
    read_type : str
        Return response as ``text``, ``read``, or ``json``.
    r_kwds : dict
        Keywords to pass to the response read function.
        It is ``{"content_type": None}`` if ``read`` is ``json``
        else an empty ``dict``.
    raise_status : bool
        Raise an exception if the response status is not 200. If
        ``False`` return ``None``.

    Returns
    -------
    tuple
        A tuple of the URL ID and the response content
    """
    async with session(url, **s_kwds) as response:
        try:
            return uid, await getattr(response, read_type)(**r_kwds)
        except UnicodeDecodeError:
            return uid, await response.text(encoding="latin1")
        except (ClientResponseError, ValueError) as ex:
            if raise_status:
                raise ServiceError(await response.text(), str(response.url)) from ex
            return uid, None


def is_jupyter_kernel():
    """Check if the code is running in a Jupyter kernel (not IPython terminal)."""
    try:
        from IPython import get_ipython

        ipython = get_ipython()
    except (ImportError, NameError):
        return False
    if ipython is None:
        return False
    return "Terminal" not in ipython.__class__.__name__


def get_event_loop() -> tuple[asyncio.AbstractEventLoop, bool]:
    """Create an event loop."""
    try:
        loop = asyncio.get_running_loop()
        new_loop = False
    except RuntimeError:
        loop = asyncio.new_event_loop()
        new_loop = True
    asyncio.set_event_loop(loop)
    if is_jupyter_kernel():
        import nest_asyncio

        nest_asyncio.apply(loop)
    return loop, new_loop


async def delete_url(
    url: StrOrURL,
    method: Literal["get", "post"],
    cache_name: Path,
    **kwargs: dict[str, Any],
) -> None:
    """Delete cached response associated with ``url``."""
    cache = SQLiteBackend(cache_name=str(cache_name))
    await cache.delete_url(url, method.upper(), **kwargs)


class BaseRetriever:
    """Base class for async retriever."""

    def __init__(
        self,
        urls: Sequence[StrOrURL],
        read_method: Literal["text", "json", "binary"],
        request_kwds: Sequence[dict[str, Any]] | None = None,
        request_method: Literal["get", "post"] = "get",
        cache_name: Path | str | None = None,
        ssl: SSLContext | bool = True,
    ) -> None:
        """Validate inputs to retrieve function."""
        ssl_cert = os.getenv("HYRIVER_SSL_CERT")
        if ssl_cert:
            self.ssl = SSLContext(PROTOCOL_TLS_CLIENT)
            self.ssl.load_verify_locations(ssl_cert)
        else:
            self.ssl = ssl

        valid_methods = ["get", "post"]
        if request_method.lower() not in valid_methods:
            raise InputValueError("method", valid_methods)
        self.request_method: Literal["get", "post"] = request_method
        valid_reads = ["binary", "json", "text"]
        if read_method not in valid_reads:
            raise InputValueError("read", valid_reads)
        self.read_method: Literal["read", "text", "json"] = (
            "read" if read_method == "binary" else read_method
        )
        self.r_kwds = {"content_type": None, "loads": json.loads} if read_method == "json" else {}

        self.url_kwds = self.generate_requests(urls, request_kwds)

        self.cache_name = create_cachefile(cache_name)

    @staticmethod
    def generate_requests(
        urls: Sequence[StrOrURL],
        request_kwds: Sequence[dict[str, Any]] | None,
    ) -> zip[tuple[int, StrOrURL, dict[str, Any]]]:
        """Generate urls and keywords."""
        if not isinstance(urls, (list, tuple)):
            raise InputTypeError("``urls``", "list of str", "[url1, ...]")

        url_id = range(len(urls))

        if request_kwds is None:
            return zip(url_id, urls, repeat({"headers": None}))

        if len(urls) != len(request_kwds):
            raise InputTypeError("urls/request_kwds", "sequences of the same size")

        valid_kwds = signature(ClientSession._request)  # pyright: ignore[reportPrivateUsage]
        not_found = [p for kwds in request_kwds for p in kwds if p not in valid_kwds.parameters]
        if not_found:
            invalids = f"request_kwds ({', '.join(not_found)})"
            raise InputValueError(invalids, list(valid_kwds.parameters))

        return zip(url_id, urls, request_kwds)



"""Download multiple files concurrently by streaming their content to disk."""

from __future__ import annotations

import asyncio
from collections.abc import Sequence
from pathlib import Path
from ssl import SSLContext
from typing import TYPE_CHECKING, Literal

import aiofiles
from aiohttp import ClientSession, ClientTimeout, TCPConnector

from async_retriever._utils import get_event_loop
from async_retriever.exceptions import DownloadError

if TYPE_CHECKING:
    from aiohttp.typedefs import StrOrURL

__all__ = ["stream_write"]
CHUNK_SIZE = 1024 * 1024  # Default chunk size of 1 MB
TIMEOUT = 10 * 60  # Timeout for requests in seconds


async def _stream_file(
    session: ClientSession,
    request_method: Literal["get", "post"],
    url: StrOrURL,
    filepath: Path,
    chunk_size: int,
) -> None:
    """Stream the response to a file."""
    async with session.request(request_method, url) as response:
        if response.status != 200:
            filepath.unlink(missing_ok=True)
            raise DownloadError(await response.text(), str(response.url))
        remote_size = int(response.headers.get("Content-Length", -1))
        if filepath.exists() and filepath.stat().st_size == remote_size:
            return

        async with aiofiles.open(filepath, "wb") as file:
            async for chunk in response.content.iter_chunked(chunk_size):
                await file.write(chunk)


async def _stream_session(
    url_file_mappings: zip[tuple[StrOrURL, Path]],
    request_method: Literal["get", "post"],
    ssl: bool | SSLContext,
    chunk_size: int,
    limit_per_host: int,
) -> None:
    """Create an async session to download files."""
    if isinstance(ssl, bool):
        verify_ssl = ssl
        ssl_context = None
    elif isinstance(ssl, SSLContext):
        verify_ssl = True
        ssl_context = ssl
    else:
        raise TypeError("`ssl` must be a boolean or SSLContext object.")

    async with ClientSession(
        connector=TCPConnector(
            verify_ssl=verify_ssl,
            ssl_context=ssl_context,
            limit_per_host=limit_per_host,
        ),
        timeout=ClientTimeout(TIMEOUT),
    ) as session:
        tasks = [
            _stream_file(session, request_method, url, filepath, chunk_size)
            for url, filepath in url_file_mappings
        ]
        await asyncio.gather(*tasks)


def stream_write(
    urls: Sequence[StrOrURL],
    file_paths: Sequence[Path] | Sequence[str],
    request_method: Literal["get", "post"] = "get",
    ssl: bool | SSLContext = True,
    chunk_size: int = CHUNK_SIZE,
    limit_per_host: int = 5,
) -> None:
    """Download multiple files concurrently by streaming their content to disk.

    Parameters
    ----------
    urls : Sequence[str]
        List of URLs to download.
    file_paths : Sequence[Path]
        List of file paths to save the downloaded content.
    request_method : {"get", "post"}, optional
        HTTP method to use (i.e., ``get`` or ``post``), by default ``get``.
    ssl : bool or ssl.SSLContext, optional
        Whether to verify SSL certificates, by default True. Also,
        an SSLContext object can be passed to customize
    chunk_size : int, optional
        Size of each chunk in bytes, by default 1 MB.
    limit_per_host : int, optional
        Maximum simultaneous connections per host, by default 5.

    Examples
    --------
    >>> import tempfile
    >>> url = "https://freetestdata.com/wp-content/uploads/2021/09/Free_Test_Data_500KB_CSV-1.csv"
    >>> with tempfile.NamedTemporaryFile(dir=".") as temp:
    ...     stream_write([url], [temp.name])
    """
    if not isinstance(urls, Sequence) or not isinstance(file_paths, Sequence):
        raise TypeError("`urls` and `file_paths` must be sequences of URLs and file paths.")
    if len(urls) != len(file_paths):
        raise TypeError("`urls` and `file_paths` must be sequences of same length.")

    file_paths = [Path(filepath) for filepath in file_paths]
    parent_dirs = {filepath.parent for filepath in file_paths}
    for parent_dir in parent_dirs:
        parent_dir.mkdir(parents=True, exist_ok=True)

    loop, is_new_loop = get_event_loop()

    try:
        loop.run_until_complete(
            _stream_session(
                zip(urls, file_paths),
                request_method,
                ssl,
                chunk_size,
                limit_per_host,
            )
        )
    finally:
        if is_new_loop:
            loop.close()
